/*
 * Copyright (C) 2010 Wouter Godefroy
 *
 * This file is part of anonip.
 *
 * Anonip is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation, either version 3 of the License, or (at your option) any later
 * version.
 *
 * Anonip is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * anonip.  If not, see <http://www.gnu.org/licenses/>.
 */


#include <sstream>
#include <iostream>
#include <fstream>
#include <string>
#include <boost/regex.hpp>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>


#ifndef RANDOM_DEVICE

#include <sys/time.h>

#endif


#include "cpan-dts/panonymizer.h"

using namespace std;

#include "config.h"

// Parse commandline

static void usage() 
{
    cout << "Usage: anonip [ -i <infile> ] [ -o <outfile> ] [ -s <seedstring> ] " << endl
	 << "              [ -m ] [ -v ] [ -V ] "
	 << endl << endl
	 << "   infile     : text file to read, default is stdin" << endl
	 << "   outfile    : text file to write, default is stdout" << endl
	 << "   seedstfing : transform seed value, default is random" << endl
	 << "   -m         : don't detect netmasks (speed up)" << endl
	 << "   -v         : be verbose" << endl
	 << "   -V         : display version and exit" << endl
	 << endl;

    exit(1);
}

static void disp_version() 
{
    cout <<  PACKAGE_TARNAME << " - (C) 2010 Wouter Godefroy" << endl;
    exit(0);
}

static string set_string_arg(stringstream& ss)
{
    string argvalue;
    ss >> argvalue;

    if(!argvalue.length())
	usage();
    
    return argvalue;
}

static void process(const string& argline) 
{
    stringstream ss;
    ss << argline;

    string arg;
    ss >> arg;
    while(arg.length() && !ss.eof()) {
	if(arg == "-V")
	    disp_version();
	else if(arg == "-v")
	    options.verbose_lvl += 1;
	else if(arg == "-o")
	    options.outfile = set_string_arg(ss);
	else if(arg == "-i")
	    options.infile = set_string_arg(ss);
	else if(arg == "-s")
	    options.seed = set_string_arg(ss);
	else if(arg == "-m")
	    options.mask = false;
	else
	    usage();
	
	arg.empty();
	ss >> arg;
    }
}



static  void process_args(int argc, char* argv[]) 
{
    if(argc < 2)
	return;

    string argline;
    for(int i = 1; i < argc; i++) 
	argline += argv[i] + string(" ");

    process(argline);
}


static string randomize_ip(PAnonymizer& pan, const string& ip) 
{
    in_addr tmp_ip;

    inet_aton(ip.c_str(), &tmp_ip);
    
    in_addr tmp_result;
    tmp_result.s_addr = pan.anonymize(tmp_ip.s_addr);
    
    return string(inet_ntoa(tmp_result));
}   

#define EXTRACT_32BITS(p) \
	((u_int32_t)((u_int32_t)*((const u_int8_t *)(p) + 0) << 24 | \
		     (u_int32_t)*((const u_int8_t *)(p) + 1) << 16 | \
		     (u_int32_t)*((const u_int8_t *)(p) + 2) << 8 | \
		     (u_int32_t)*((const u_int8_t *)(p) + 3)))


static bool is_netmask(const string& ip) 
{
    in_addr tmp_ip;

    inet_aton(ip.c_str(), &tmp_ip);

    uint32_t ip_n = EXTRACT_32BITS(&tmp_ip.s_addr);

    bool saw_one = false;
    for(int i=0; i<32; i++) {
	if((ip_n & 1))
	    saw_one = true;
	else
	    if(saw_one)
		return false;
	
	ip_n >>= 1;
    }	

    return true;
}


static void convert_file(istream& in, ostream& out, PAnonymizer& pan) 
{
    static const boost::regex re_match_all_ip( "((25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9][0-9]|[0-9])\\.(25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9][0-9]|[0-9])\\.(25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9][0-9]|[0-9])\\.(25[0-5]|2[0-4][0-9]|1[0-9][0-9]|[1-9][0-9]|[0-9]))");
    static const boost::sregex_iterator iterend;

    while(!in.eof()) {
	string line;
	getline(in, line);

	list<string> iplist;
	for(boost::sregex_iterator m1(line.begin(), line.end(), re_match_all_ip);m1 != iterend; m1++) 
	    iplist.push_back((*m1)[0]);

	iplist.sort();
	iplist.unique();
	
	for(list<string>::const_iterator ipl_iter = iplist.begin(); ipl_iter != iplist.end(); ipl_iter++)  {
	    if(options.mask && is_netmask(*ipl_iter))
		continue;

	    string new_ip = randomize_ip(pan, *ipl_iter);
	    
	    if(options.verbose_lvl)
		cerr << *ipl_iter << " -> " << new_ip << endl;
	    
	    string search_ip = boost::regex_replace(*ipl_iter, boost::regex("\\."), "\\\\.", boost::regex_constants::match_any & boost::regex_constants::format_literal);
	    line = boost::regex_replace(line, boost::regex(search_ip), new_ip, boost::regex_constants::match_any & boost::regex_constants::format_literal);
	}
	
	out << line << endl;
      }
}

int main(int argc, char* argv[])
{

    process_args(argc, argv);
    
    // open files
    ifstream finfile;
    if(!options.infile.empty()) {
	finfile.open(options.infile.c_str(), ios::in);

	if(!finfile) {
	    perror(options.infile.c_str());
	    exit(2);
	}
    }

    ofstream foutfile;
    if(!options.outfile.empty()) {
	foutfile.open(options.outfile.c_str(), ios::out);

	if(!foutfile) {
	    perror(options.outfile.c_str());
	    exit(3);
	}
    }

    // TODO: Provide your own 256-bit key here
    unsigned char seed_key[32];
    if(options.seed.empty()) {
#ifdef RANDOM_DEVICE
	ifstream random(RANDOM_DEVICE);
	if(!random) {
	    perror(RANDOM_DEVICE);	    
	    exit(3);
	}
	
	random.get((char*) seed_key, 32);
#else
	timeval tv;
	gettimeofday(&tv, 0);
	srand((tv.tv_sec ^ tv.tv_usec) ^ getpid());
	for(int i=0;i<32;i++)
	    seed_key[i] = rand() % 256;
#endif
	
    } else {
	if(options.seed.length() > 32 ) {
	    cerr << "Error: Seed should be no longer than 32 characters!" << endl;
	    exit (3);
	}
		
	int mul = 32 / options.seed.length();
	int mod = 32 % options.seed.length();

	string seed;
	for(int i=0; i< mul; i++) seed += options.seed;
	seed += string(options.seed,0,mod);

	seed.copy((char*)seed_key, 32);
    }
        
    PAnonymizer ip_anonymizer(seed_key);
    
    convert_file(options.infile.empty()  ? cin : finfile,
		 options.outfile.empty() ? cout : foutfile,
		 ip_anonymizer);

    finfile.close();
    foutfile.close();
    
    return 0;
}

